@@ -5,10 +5,12 @@ class TaggedObject:
     A class that takes arbitrary tags.
     """
 
-    __slots__ = ('tags', )
+    __slots__ = ('idx', 'tags', '_hash', )
 
-    def __init__(self, **kwargs):
+    def __init__(self, idx, **kwargs):
         self.tags = { }
+        self.idx = idx
+        self._hash = None
         if kwargs:
             self.initialize_tags(kwargs)
 
@@ -21,3 +23,11 @@ class TaggedObject:
             return self.tags[item]
         except KeyError:
             return super(TaggedObject, self).__getattribute__(item)
+
+    def __hash__(self):
+        if self._hash is None:
+            self._hash = self._hash_core()
+        return self._hash
+
+    def _hash_core(self):
+        raise NotImplementedError()
