@@ -1,5 +1,10 @@
+from typing import Optional, TYPE_CHECKING
 
 from .tagged_object import TaggedObject
+from .expression import Expression
+
+if TYPE_CHECKING:
+    from angr.calling_conventions import SimCC
 
 
 class Statement(TaggedObject):
@@ -7,12 +12,10 @@ class Statement(TaggedObject):
     The base class of all AIL statements.
     """
 
-    __slots__ = ('idx', )
+    __slots__ = ()
 
     def __init__(self, idx, **kwargs):
-        super(Statement, self).__init__(**kwargs)
-
-        self.idx = idx
+        super(Statement, self).__init__(idx, **kwargs)
 
     def __repr__(self):
         raise NotImplementedError()
@@ -20,9 +23,6 @@ class Statement(TaggedObject):
     def __str__(self):
         raise NotImplementedError()
 
-    def __hash__(self):
-        raise NotImplementedError()
-
     def replace(self, old_expr, new_expr):
         raise NotImplementedError()
 
@@ -46,7 +46,9 @@ class Assignment(Statement):
                self.dst == other.dst and \
                self.src == other.src
 
-    def __hash__(self):
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
         return hash((Assignment, self.idx, self.dst, self.src))
 
     def __repr__(self):
@@ -57,8 +59,17 @@ class Assignment(Statement):
 
     def replace(self, old_expr, new_expr):
 
-        r_dst, replaced_dst = self.dst.replace(old_expr, new_expr)
-        r_src, replaced_src = self.src.replace(old_expr, new_expr)
+        if self.dst.likes(old_expr):
+            r_dst = True
+            replaced_dst = new_expr
+        else:
+            r_dst, replaced_dst = self.dst.replace(old_expr, new_expr)
+
+        if self.src.likes(old_expr):
+            r_src = True
+            replaced_src = new_expr
+        else:
+            r_src, replaced_src = self.src.replace(old_expr, new_expr)
 
         if r_dst or r_src:
             return True, Assignment(self.idx, replaced_dst, replaced_src, **self.tags)
@@ -90,7 +101,9 @@ class Store(Statement):
                self.guard == other.guard and \
                self.endness == other.endness
 
-    def __hash__(self):
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
         return hash((Store, self.idx, self.addr, self.data, self.size, self.endness, self.guard))
 
     def __repr__(self):
@@ -106,8 +119,18 @@ class Store(Statement):
                                         "" if self.guard is None else "[%s]" % self.guard)
 
     def replace(self, old_expr, new_expr):
-        r_addr, replaced_addr = self.addr.replace(old_expr, new_expr)
-        r_data, replaced_data = self.data.replace(old_expr, new_expr)
+        if self.addr.likes(old_expr):
+            r_addr = True
+            replaced_addr = new_expr
+        else:
+            r_addr, replaced_addr = self.addr.replace(old_expr, new_expr)
+
+        if self.data.likes(old_expr):
+            r_data = True
+            replaced_data = new_expr
+        else:
+            r_data, replaced_data = self.data.replace(old_expr, new_expr)
+
         if self.guard is not None:
             r_guard, replaced_guard = self.guard.replace(old_expr, new_expr)
         else:
@@ -134,7 +157,9 @@ class Jump(Statement):
                self.idx == other.idx and \
                self.target == other.target
 
-    def __hash__(self):
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
         return hash((Jump, self.idx, self.target))
 
     def __repr__(self):
@@ -151,6 +176,13 @@ class Jump(Statement):
         else:
             return False, self
 
+    def copy(self):
+        return Jump(
+            self.idx,
+            self.target,
+            **self.tags,
+        )
+
 
 class ConditionalJump(Statement):
 
@@ -170,7 +202,9 @@ class ConditionalJump(Statement):
                self.true_target == other.true_target and \
                self.false_target == other.false_target
 
-    def __hash__(self):
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
         return hash((ConditionalJump, self.idx, self.condition, self.true_target, self.false_target))
 
     def __repr__(self):
@@ -203,12 +237,17 @@ class ConditionalJump(Statement):
             return False, self
 
 
-class Call(Statement):
+class Call(Expression, Statement):
+    """
+    Call is both an expression and a statement. The return expression of a call is defined as the ret_expr if and only
+    if the callee function has one return expression.
+    """
 
     __slots__ = ('target', 'calling_convention', 'prototype', 'args', 'ret_expr', )
 
-    def __init__(self, idx, target, calling_convention=None, prototype=None, args=None, ret_expr=None, **kwargs):
-        super(Call, self).__init__(idx, **kwargs)
+    def __init__(self, idx, target, calling_convention: Optional['SimCC']=None, prototype=None, args=None, ret_expr=None,
+                 **kwargs):
+        super().__init__(idx, target.depth + 1,**kwargs)
 
         self.target = target
         self.calling_convention = calling_convention
@@ -225,7 +264,9 @@ class Call(Statement):
                self.args == other.args and \
                self.ret_expr == other.ret_expr
 
-    def __hash__(self):
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
         return hash((Call, self.idx, self.target))
 
     def __repr__(self):
@@ -235,7 +276,11 @@ class Call(Statement):
 
         cc = "Unknown CC" if self.calling_convention is None else "%s" % self.calling_convention
         if self.args is None:
-            s = ("%s" % cc) if self.prototype is None else "%s: %s" % (self.calling_convention, self.calling_convention.arg_locs())
+            if self.calling_convention is not None:
+                s = ("%s" % cc) if self.prototype is None else "%s: %s" % (self.calling_convention,
+                                                                           self.calling_convention.arg_locs())
+            else:
+                s = ("%s" % cc) if self.prototype is None else repr(self.prototype)
         else:
             s = ("%s" % cc) if self.prototype is None else "%s: %s" % (self.calling_convention, self.args)
 
@@ -244,6 +289,22 @@ class Call(Statement):
             s
         )
 
+    @property
+    def bits(self):
+        return self.ret_expr.bits
+
+    @property
+    def size(self):
+        return self.bits // 8
+
+    @property
+    def verbose_op(self):
+        return "call"
+
+    @property
+    def op(self):
+        return "call"
+
     def replace(self, old_expr, new_expr):
         r0, replaced_target = self.target.replace(old_expr, new_expr)
 
@@ -294,7 +355,7 @@ class Return(Statement):
         super().__init__(idx, **kwargs)
 
         self.target = target
-        self.ret_exprs = list(ret_exprs) if not isinstance(ret_exprs, list) else [ ]
+        self.ret_exprs = ret_exprs if isinstance(ret_exprs, list) else list(ret_exprs)
 
     def __eq__(self, other):
         return type(other) is Return and \
@@ -302,8 +363,10 @@ class Return(Statement):
                 self.target == other.target and \
                 self.ret_exprs == other.ret_exprs
 
-    def __hash__(self):
-        return hash((self.Return, self.idx, self.target, tuple(self.ret_exprs)))
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return hash((Return, self.idx, self.target, tuple(self.ret_exprs)))
 
     def __repr__(self):
         return "Return to %r (%r)" % (self.target, ",".join(self.ret_exprs))
@@ -363,7 +426,7 @@ class DirtyStatement(Statement):
         super(DirtyStatement, self).__init__(idx, **kwargs)
         self.dirty_stmt = dirty_stmt
 
-    def __hash__(self):
+    def _hash_core(self):
         return hash((DirtyStatement, self.dirty_stmt))
 
     def __repr__(self):
